# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.utils.py3 import httplib
from libcloud.dns.drivers.luadns import LuadnsDNSDriver
from libcloud.test import MockHttp
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.test.secrets import DNS_PARAMS_LUADNS
from libcloud.dns.types import ZoneDoesNotExistError, ZoneAlreadyExistsError
from libcloud.dns.types import RecordDoesNotExistError
from libcloud.dns.types import RecordType
from libcloud.dns.base import Zone, Record


class LuadnsTests(unittest.TestCase):

    def setUp(self):
        LuadnsMockHttp.type = None
        LuadnsDNSDriver.connectionCls.conn_class = LuadnsMockHttp
        self.driver = LuadnsDNSDriver(*DNS_PARAMS_LUADNS)
        self.test_zone = Zone(id='11', type='master', ttl=None,
                              domain='example.com', extra={},
                              driver=self.driver)
        self.test_record = Record(id='13', type=RecordType.A,
                                  name='example.com', zone=self.test_zone,
                                  data='127.0.0.1', driver=self, extra={})

    def assertHasKeys(self, dictionary, keys):
        for key in keys:
            self.assertTrue(key in dictionary,
                            'key "%s" not in dictionary' % key)

    def test_list_zones_empty(self):
        LuadnsMockHttp.type = 'EMPTY_ZONES_LIST'
        zones = self.driver.list_zones()

        self.assertEqual(zones, [])

    def test_list_zones_success(self):
        zones = self.driver.list_zones()

        self.assertEqual(len(zones), 2)

        zone = zones[0]
        self.assertEqual(zone.id, '1')
        self.assertEqual(zone.domain, 'example.com')
        self.assertIsNone(zone.type)
        self.assertEqual(zone.driver, self.driver)
        self.assertIsNone(zone.ttl)

        second_zone = zones[1]
        self.assertEqual(second_zone.id, '2')
        self.assertEqual(second_zone.domain, 'example.net')
        self.assertIsNone(second_zone.type)
        self.assertEqual(second_zone.driver, self.driver)
        self.assertIsNone(second_zone.ttl)

    def test_get_zone_zone_does_not_exist(self):
        LuadnsMockHttp.type = 'ZONE_DOES_NOT_EXIST'
        try:
            self.driver.get_zone(zone_id='13')
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, '13')
        else:
            self.fail('Exception was not thrown')

    def test_get_zone_success(self):
        LuadnsMockHttp.type = 'GET_ZONE_SUCCESS'
        zone = self.driver.get_zone(zone_id='31')

        self.assertEqual(zone.id, '31')
        self.assertEqual(zone.domain, 'example.org')
        self.assertIsNone(zone.type)
        self.assertIsNone(zone.ttl)
        self.assertEqual(zone.driver, self.driver)

    def test_delete_zone_success(self):
        LuadnsMockHttp.type = 'DELETE_ZONE_SUCCESS'
        zone = self.test_zone
        status = self.driver.delete_zone(zone=zone)

        self.assertEqual(status, True)

    def test_delete_zone_zone_does_not_exist(self):
        LuadnsMockHttp.type = 'DELETE_ZONE_ZONE_DOES_NOT_EXIST'
        zone = self.test_zone
        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, '11')
        else:
            self.fail('Exception was not thrown')

    def test_create_zone_success(self):
        LuadnsMockHttp.type = 'CREATE_ZONE_SUCCESS'
        zone = self.driver.create_zone(domain='example.org')

        self.assertEqual(zone.id, '3')
        self.assertEqual(zone.domain, 'example.org')
        self.assertIsNone(zone.type)
        self.assertIsNone(zone.ttl)
        self.assertEqual(zone.driver, self.driver)

    def test_create_zone_zone_zone_already_exists(self):
        LuadnsMockHttp.type = 'CREATE_ZONE_ZONE_ALREADY_EXISTS'
        try:
            self.driver.create_zone(domain='test.com')
        except ZoneAlreadyExistsError as e:
            self.assertEqual(e.zone_id, 'test.com')
        else:
            self.fail('Exception was not thrown')

    def test_list_records_empty(self):
        LuadnsMockHttp.type = 'EMPTY_RECORDS_LIST'
        zone = self.test_zone
        records = self.driver.list_records(zone=zone)

        self.assertEqual(records, [])

    def test_list_records_success(self):
        LuadnsMockHttp.type = 'LIST_RECORDS_SUCCESS'
        zone = self.test_zone
        records = self.driver.list_records(zone=zone)

        self.assertEqual(len(records), 2)

        record = records[0]
        self.assertEqual(record.id, '6683')
        self.assertEqual(record.type, 'NS')
        self.assertEqual(record.name, 'example.org.')
        self.assertEqual(record.data, 'b.ns.luadns.net.')
        self.assertEqual(record.zone, self.test_zone)
        self.assertEqual(record.zone.id, '11')

        second_record = records[1]
        self.assertEqual(second_record.id, '6684')
        self.assertEqual(second_record.type, 'NS')
        self.assertEqual(second_record.name, 'example.org.')
        self.assertEqual(second_record.data, 'a.ns.luadns.net.')
        self.assertEqual(second_record.zone, self.test_zone)

    def test_get_record_record_does_not_exist(self):
        LuadnsMockHttp.type = 'GET_RECORD_RECORD_DOES_NOT_EXIST'
        try:
            self.driver.get_record(zone_id='31', record_id='31')
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, '31')
        else:
            self.fail('Exception was not thrown')

    def test_get_record_success(self):
        LuadnsMockHttp.type = 'GET_RECORD_SUCCESS'
        record = self.driver.get_record(zone_id='31', record_id='31')

        self.assertEqual(record.id, '31')
        self.assertEqual(record.type, 'MX')
        self.assertEqual(record.name, 'example.com.')
        self.assertEqual(record.data, '10 mail.example.com.')

    def test_delete_record_success(self):
        LuadnsMockHttp.type = 'DELETE_RECORD_SUCCESS'
        record = self.test_record
        status = self.driver.delete_record(record=record)

        self.assertEqual(status, True)

    def test_delete_record_RECORD_DOES_NOT_EXIST_ERROR(self):
        LuadnsMockHttp.type = 'DELETE_RECORD_RECORD_DOES_NOT_EXIST'
        record = self.test_record
        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, '13')
        else:
            self.fail('Exception was not thrown')

    def test_create_record_success(self):
        LuadnsMockHttp.type = 'CREATE_RECORD_SUCCESS'
        record = self.driver.create_record(name='test.com.',
                                           zone=self.test_zone,
                                           type='A',
                                           data='127.0.0.1',
                                           extra={'ttl': 13})
        self.assertEqual(record.id, '31')
        self.assertEqual(record.name, 'test.com.')
        self.assertEqual(record.data, '127.0.0.1')
        self.assertIsNone(record.ttl)

    def test_record_already_exists_error(self):
        pass


class LuadnsMockHttp(MockHttp):
    fixtures = DNSFileFixtures('luadns')

    def _v1_zones(self, method, url, body, headers):
        body = self.fixtures.load('zones_list.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_EMPTY_ZONES_LIST(self, method, url, body,
                                   headers):
        body = self.fixtures.load('empty_zones_list.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_13_ZONE_DOES_NOT_EXIST(self, method, url,
                                         body, headers):
        body = self.fixtures.load('zone_does_not_exist.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_GET_ZONE_SUCCESS(self, method, url,
                                      body, headers):
        body = self.fixtures.load('get_zone.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_DELETE_ZONE_SUCCESS(self, method, url,
                                         body, headers):
        body = self.fixtures.load('delete_zone_success.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_DELETE_ZONE_ZONE_DOES_NOT_EXIST(
            self, method, url, body, headers):
        body = self.fixtures.load('zone_does_not_exist.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_CREATE_ZONE_SUCCESS(self, method, url,
                                      body, headers):
        body = self.fixtures.load('create_zone_success.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_CREATE_ZONE_ZONE_ALREADY_EXISTS(
            self, method, url, body, headers):
        body = self.fixtures.load('zone_already_exists.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_EMPTY_RECORDS_LIST(self, method, url, body,
                                                headers):
        body = self.fixtures.load('empty_records_list.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_LIST_RECORDS_SUCCESS(self, method, url,
                                                  body, headers):
        body = self.fixtures.load('records_list.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_records_31_GET_RECORD_RECORD_DOES_NOT_EXIST(
            self, method, url, body, headers):
        body = self.fixtures.load('record_does_not_exist.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_GET_RECORD_RECORD_DOES_NOT_EXIST(
            self, method, url, body, headers):

        body = self.fixtures.load('get_zone.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_GET_RECORD_SUCCESS(self, method, url,
                                        body, headers):
        body = self.fixtures.load('get_zone.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_records_31_GET_RECORD_SUCCESS(self, method, url,
                                                   body, headers):
        body = self.fixtures.load('get_record.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_13_DELETE_RECORD_SUCCESS(self, method, url,
                                                      body, headers):
        body = self.fixtures.load('delete_record_success.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_13_DELETE_RECORD_RECORD_DOES_NOT_EXIST(
            self, method, url, body, headers):
        body = self.fixtures.load('record_does_not_exist.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_CREATE_RECORD_SUCCESS(self, method, url,
                                                   body, headers):
        body = self.fixtures.load('create_record_success.json')

        return httplib.OK, body, {}, httplib.responses[httplib.OK]


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

import requests
import requests_mock

from libcloud.common.base import XmlResponse, JsonResponse, Connection
from libcloud.common.types import MalformedResponseError
from libcloud.http import LibcloudConnection


class ResponseClassesTests(unittest.TestCase):
    def setUp(self):
        self.mock_connection = LibcloudConnection(host='mock.com', port=80)
        self.mock_connection.driver = None

    def test_XmlResponse_class(self):
        with requests_mock.mock() as m:
            m.register_uri('GET', 'mock://test.com/2', text='<foo>bar</foo>')
            response_obj = requests.get('mock://test.com/2')
            response = XmlResponse(response=response_obj,
                                   connection=self.mock_connection)

        parsed = response.parse_body()
        self.assertEqual(parsed.tag, 'foo')
        self.assertEqual(parsed.text, 'bar')

    def test_XmlResponse_class_malformed_response(self):
        with requests_mock.mock() as m:
            m.register_uri('GET', 'mock://test.com/', text='<foo>')
            response_obj = requests.get('mock://test.com/')
            try:
                XmlResponse(response=response_obj,
                            connection=self.mock_connection)
            except MalformedResponseError:
                pass
            else:
                self.fail('Exception was not thrown')

    def test_XmlResponse_class_zero_length_body_strip(self):
        with requests_mock.mock() as m:
            m.register_uri('GET', 'mock://test.com/', text=' ')
            response_obj = requests.get('mock://test.com/')
            response = XmlResponse(response=response_obj,
                                   connection=self.mock_connection)

        parsed = response.parse_body()
        self.assertEqual(parsed, '')

    def test_JsonResponse_class_success(self):
        with requests_mock.mock() as m:
            m.register_uri('GET', 'mock://test.com/', text='{"foo": "bar"}')
            response_obj = requests.get('mock://test.com/')
            response = JsonResponse(response=response_obj,
                                    connection=self.mock_connection)

        parsed = response.parse_body()
        self.assertEqual(parsed, {'foo': 'bar'})

    def test_JsonResponse_class_malformed_response(self):
        with requests_mock.mock() as m:
            m.register_uri('GET', 'mock://test.com/', text='{"foo": "bar"')
            response_obj = requests.get('mock://test.com/')
            try:
                JsonResponse(response=response_obj,
                             connection=self.mock_connection)
            except MalformedResponseError:
                pass
            else:
                self.fail('Exception was not thrown')

    def test_JsonResponse_class_zero_length_body_strip(self):
        with requests_mock.mock() as m:
            m.register_uri('GET', 'mock://test.com/', text=' ')
            response_obj = requests.get('mock://test.com/')
            response = JsonResponse(response=response_obj,
                                    connection=self.mock_connection)

        parsed = response.parse_body()
        self.assertEqual(parsed, '')

    def test_RawResponse_class_read_method(self):
        """
        Test that the RawResponse class includes a response
        property which exhibits the same properties and methods
        as httplib.HTTPResponse for backward compat <1.5.0
        """
        TEST_DATA = '1234abcd'

        conn = Connection(host='mock.com', port=80, secure=False)
        conn.connect()

        with requests_mock.Mocker() as m:
            m.register_uri('GET', 'http://mock.com/raw_data', text=TEST_DATA,
                           headers={'test': 'value'})
            response = conn.request('/raw_data', raw=True)
        data = response.response.read()
        self.assertEqual(data, TEST_DATA)

        header_value = response.response.getheader('test')
        self.assertEqual(header_value, 'value')

        headers = response.response.getheaders()
        self.assertEqual(headers, [('test', 'value')])

        self.assertEqual(response.response.status, 200)

if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.utils.py3 import httplib

from libcloud.dns.types import RecordType
from libcloud.dns.drivers.dnsimple import DNSimpleDNSDriver

from libcloud.test import MockHttp
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.test.secrets import DNS_PARAMS_DNSIMPLE


class DNSimpleDNSTests(unittest.TestCase):
    def setUp(self):
        DNSimpleDNSDriver.connectionCls.conn_class = DNSimpleDNSMockHttp
        DNSimpleDNSMockHttp.type = None
        self.driver = DNSimpleDNSDriver(*DNS_PARAMS_DNSIMPLE)

    def assertHasKeys(self, dictionary, keys):
        for key in keys:
            self.assertTrue(key in dictionary, 'key "%s" not in dictionary' %
                            (key))

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 15)
        self.assertTrue(RecordType.A in record_types)
        self.assertTrue(RecordType.AAAA in record_types)
        self.assertTrue(RecordType.ALIAS in record_types)
        self.assertTrue(RecordType.CNAME in record_types)
        self.assertTrue(RecordType.HINFO in record_types)
        self.assertTrue(RecordType.MX in record_types)
        self.assertTrue(RecordType.NAPTR in record_types)
        self.assertTrue(RecordType.NS in record_types)
        self.assertTrue('POOL' in record_types)
        self.assertTrue(RecordType.SPF in record_types)
        self.assertTrue(RecordType.SOA in record_types)
        self.assertTrue(RecordType.SRV in record_types)
        self.assertTrue(RecordType.SSHFP in record_types)
        self.assertTrue(RecordType.TXT in record_types)
        self.assertTrue(RecordType.URL in record_types)

    def test_list_zones_success(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)

        zone1 = zones[0]
        self.assertEqual(zone1.id, '1')
        self.assertEqual(zone1.type, 'master')
        self.assertEqual(zone1.domain, 'example.com')
        self.assertEqual(zone1.ttl, 3600)
        self.assertHasKeys(zone1.extra, ['registrant_id', 'user_id',
                                         'unicode_name', 'token', 'state',
                                         'language', 'lockable', 'auto_renew',
                                         'whois_protected', 'record_count',
                                         'service_count', 'expires_on',
                                         'created_at', 'updated_at'])

        zone2 = zones[1]
        self.assertEqual(zone2.id, '2')
        self.assertEqual(zone2.type, 'master')
        self.assertEqual(zone2.domain, 'example.com')
        self.assertEqual(zone2.ttl, 3600)
        self.assertHasKeys(zone2.extra, ['registrant_id', 'user_id',
                                         'unicode_name', 'token', 'state',
                                         'language', 'lockable', 'auto_renew',
                                         'whois_protected', 'record_count',
                                         'service_count', 'expires_on',
                                         'created_at', 'updated_at'])

    def test_list_records_success(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 3)

        record1 = records[0]
        self.assertEqual(record1.id, '31')
        self.assertEqual(record1.name, '')
        self.assertEqual(record1.type, RecordType.A)
        self.assertEqual(record1.data, '1.2.3.4')
        self.assertHasKeys(record1.extra, ['ttl', 'created_at', 'updated_at',
                                           'domain_id', 'priority'])

        record2 = records[1]
        self.assertEqual(record2.id, '2')
        self.assertEqual(record2.name, 'www')
        self.assertEqual(record2.type, RecordType.CNAME)
        self.assertEqual(record2.data, 'example.com')
        self.assertHasKeys(record2.extra, ['ttl', 'created_at', 'updated_at',
                                           'domain_id', 'priority'])

        record3 = records[2]
        self.assertEqual(record3.id, '32')
        self.assertEqual(record3.name, '')
        self.assertEqual(record3.type, RecordType.MX)
        self.assertEqual(record3.data, 'mail.example.com')
        self.assertHasKeys(record3.extra, ['ttl', 'created_at', 'updated_at',
                                           'domain_id', 'priority'])

    def test_get_zone_success(self):
        zone1 = self.driver.get_zone(zone_id='1')
        self.assertEqual(zone1.id, '1')
        self.assertEqual(zone1.type, 'master')
        self.assertEqual(zone1.domain, 'example.com')
        self.assertEqual(zone1.ttl, 3600)
        self.assertHasKeys(zone1.extra, ['registrant_id', 'user_id',
                                         'unicode_name', 'token', 'state',
                                         'language', 'lockable', 'auto_renew',
                                         'whois_protected', 'record_count',
                                         'service_count', 'expires_on',
                                         'created_at', 'updated_at'])

    def test_get_record_success(self):
        record = self.driver.get_record(zone_id='1',
                                        record_id='123')
        self.assertEqual(record.id, '123')
        self.assertEqual(record.name, 'www')
        self.assertEqual(record.type, RecordType.CNAME)
        self.assertEqual(record.data, 'example.com')
        self.assertHasKeys(record.extra, ['ttl', 'created_at', 'updated_at',
                                          'domain_id', 'priority'])

    def test_create_zone_success(self):
        DNSimpleDNSMockHttp.type = 'CREATE'
        zone = self.driver.create_zone(domain='example.com')
        self.assertEqual(zone.id, '1')
        self.assertEqual(zone.domain, 'example.com')
        self.assertEqual(zone.ttl, 3600)
        self.assertEqual(zone.type, 'master')
        self.assertHasKeys(zone.extra, ['registrant_id', 'user_id',
                                        'unicode_name', 'token', 'state',
                                        'language', 'lockable', 'auto_renew',
                                        'whois_protected', 'record_count',
                                        'service_count', 'expires_on',
                                        'created_at', 'updated_at'])

    def test_create_record_success(self):
        zone = self.driver.list_zones()[0]
        DNSimpleDNSMockHttp.type = 'CREATE'
        record = self.driver.create_record(name='domain4', zone=zone,
                                           type=RecordType.MX,
                                           data='mail.example.com')
        self.assertEqual(record.id, '172')
        self.assertEqual(record.name, '')
        self.assertEqual(record.type, RecordType.MX)
        self.assertEqual(record.data, 'mail.example.com')
        self.assertHasKeys(record.extra, ['ttl', 'created_at', 'updated_at',
                                          'domain_id', 'priority'])

    def test_update_record_success(self):
        record = self.driver.get_record(zone_id='1',
                                        record_id='123')
        DNSimpleDNSMockHttp.type = 'UPDATE'
        extra = {'ttl': 4500}
        record1 = self.driver.update_record(record=record, name='www',
                                            type=record.type,
                                            data='updated.com',
                                            extra=extra)
        self.assertEqual(record.data, 'example.com')
        self.assertEqual(record.extra.get('ttl'), 3600)
        self.assertEqual(record1.data, 'updated.com')
        self.assertEqual(record1.extra.get('ttl'), 4500)

    def test_delete_zone_success(self):
        zone = self.driver.list_zones()[0]
        DNSimpleDNSMockHttp.type = 'DELETE_200'
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_success_future_implementation(self):
        zone = self.driver.list_zones()[0]
        DNSimpleDNSMockHttp.type = 'DELETE_204'
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_record_success(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 3)
        record = records[1]
        DNSimpleDNSMockHttp.type = 'DELETE_200'
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_success_future_implementation(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 3)
        record = records[1]
        DNSimpleDNSMockHttp.type = 'DELETE_204'
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)


class DNSimpleDNSMockHttp(MockHttp):
    fixtures = DNSFileFixtures('dnsimple')

    def _v1_domains(self, method, url, body, headers):
        body = self.fixtures.load('list_domains.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_CREATE(self, method, url, body, headers):
        body = self.fixtures.load('create_domain.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1(self, method, url, body, headers):
        body = self.fixtures.load('get_domain.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records(self, method, url, body, headers):
        body = self.fixtures.load('list_domain_records.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records_123(self, method, url, body, headers):
        body = self.fixtures.load('get_domain_record.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records_CREATE(self, method, url, body, headers):
        body = self.fixtures.load('create_domain_record.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records_123_UPDATE(self, method, url, body, headers):
        body = self.fixtures.load('update_domain_record.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_DELETE_200(self, method, url, body, headers):
        return (httplib.OK, '', {}, httplib.responses[httplib.OK])

    def _v1_domains_1_DELETE_204(self, method, url, body, headers):
        return (httplib.OK, '', {}, httplib.responses[httplib.NO_CONTENT])

    def _v1_domains_1_records_2_DELETE_200(self, method, url, body, headers):
        return (httplib.OK, '', {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records_2_DELETE_204(self, method, url, body, headers):
        return (httplib.OK, '', {}, httplib.responses[httplib.NO_CONTENT])


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from datetime import datetime
from libcloud.utils.iso8601 import UTC

from libcloud.utils.py3 import httplib
from libcloud.compute.drivers.gridscale import GridscaleNodeDriver
from libcloud.compute.base import NodeSize

from libcloud.test import LibcloudTestCase, MockHttp
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.test.secrets import GRIDSCALE_PARAMS


class Gridscale_Tests(LibcloudTestCase):
    def setUp(self):
        GridscaleNodeDriver.connectionCls.conn_class = GridscaleMockHttp
        GridscaleMockHttp.type = None
        self.driver = GridscaleNodeDriver(*GRIDSCALE_PARAMS)

    def test_create_node_success(self):
        image = self.driver.list_images()[0]
        size = NodeSize(id=0, name='test', bandwidth=0, price=0, ram=10240,
                        driver=self.driver, disk=50, extra={'cores': 2})
        location = self.driver.list_locations()[0]
        sshkey = ["b1682d3a-1869-4bdc-8ffe-e74a261d300c"]
        GridscaleMockHttp.type = 'POST'
        node = self.driver.create_node(name='test', size=size, image=image,
                                       location=location, ex_ssh_key_ids=sshkey)

        self.assertEqual(node.name, 'test.test')
        self.assertEqual(node.public_ips, ['185.102.95.236', '2a06:2380:0:1::211'])

    def test_create_image_success(self):
        node = self.driver.list_nodes()[0]
        GridscaleMockHttp.type = 'POST'
        image = self.driver.create_image(node, 'test.test')
        self.assertTrue(image)

    def test_list_nodes_success(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].name, 'test.test')
        self.assertEqual(nodes[0].public_ips, ['185.102.95.236', '2a06:2380:0:1::211'])
        self.assertEqual(nodes[0].extra['cores'], 2)
        self.assertEqual(nodes[0].extra['memory'], 10240)

    def test_list_locations_success(self):
        locations = self.driver.list_locations()
        self.assertTrue(len(locations) >= 1)

    def test_list_volumes(self):
        volumes = self.driver.list_volumes()
        self.assertEqual(len(volumes), 1)
        volume = volumes[0]
        self.assertEqual(volume.id, "e66bb753-4a03-4ee2-a069-a601f393c9ee")
        self.assertEqual(volume.name, "linux")
        self.assertEqual(volume.size, 50)
        self.assertEqual(volume.driver, self.driver)

    def test_list_images_success(self):
        images = self.driver.list_images()
        self.assertTrue(len(images) >= 1)

        image = images[0]
        self.assertTrue(image.id is not None)
        self.assertTrue(image.name is not None)

    def test_list_key_pairs(self):
        keys = self.driver.list_key_pairs()
        self.assertEqual(len(keys), 2)
        self.assertEqual(keys[0].name, 'karl')
        self.assertEqual(keys[0].public_key, "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC")

    def test_list_volume_snapshots(self):
        volume = self.driver.list_volumes()[0]
        snapshots = self.driver.list_volume_snapshots(volume)

        self.assertEqual(len(snapshots), 1)
        self.assertEqual(snapshots[0].id, "d755de62-4d75-4d61-addd-a5c9743a5deb")

    def test_list_volumes_empty(self):
        GridscaleMockHttp.type = 'EMPTY'
        volumes = self.driver.list_volumes()
        self.assertEqual(len(volumes), 0)

    def test_ex_list_networks(self):
        networks = self.driver.ex_list_networks()[0]
        self.assertEqual(networks.id, "1196529b-a8de-417f")

    def test_ex_list_ips(self):
        ip = self.driver.ex_list_ips()[0]
        self.assertEqual(ip.id, "56b8d161-325b-4fd4")

    def test_ex_destroy_ip(self):
        ip = self.driver.ex_list_ips()[0]

        GridscaleMockHttp.type = 'DELETE'
        self.assertTrue(self.driver.ex_destroy_ip(ip))

    def test_ex_destroy_network(self):
        network = self.driver.ex_list_networks()[0]

        GridscaleMockHttp.type = 'DELETE'
        self.assertTrue(self.driver.ex_destroy_network(network))

    def test_destroy_node_success(self):
        # Regular destroy
        node = self.driver.list_nodes()[0]
        GridscaleMockHttp.type = 'DELETE'

        res = self.driver.destroy_node(node)
        self.assertTrue(res)

        # Destroy associated resources
        res = self.driver.destroy_node(node, ex_destroy_associated_resources=True)
        self.assertTrue(res)

    def test_destroy_volume(self):
        volume = self.driver.list_volumes()[0]
        GridscaleMockHttp.type = 'DELETE'
        res = self.driver.destroy_volume(volume)
        self.assertTrue(res)

    def test_destroy_volume_snapshot(self):
        volume = self.driver.list_volumes()[0]
        snapshot = self.driver.list_volume_snapshots(volume)[0]
        GridscaleMockHttp.type = 'DELETE'
        res = self.driver.destroy_volume_snapshot(snapshot)
        self.assertTrue(res)

    def test_get_image_success(self):
        image = self.driver.get_image("12345")
        self.assertEqual(image.id, "12345")

    def test_list_nodes_fills_created_datetime(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(nodes[0].created_at, datetime(2019, 6, 7, 12, 56, 44, tzinfo=UTC))

    def test_ex_list_volumes_for_node(self):
        node = self.driver.list_nodes()[0]
        volumes = self.driver.ex_list_volumes_for_node(node=node)
        self.assertEqual(len(volumes), 1)
        self.assertEqual(volumes[0].size, 50)

    def test_ex_list_ips_for_node(self):
        node = self.driver.list_nodes()[0]
        ips = self.driver.ex_list_ips_for_node(node=node)
        self.assertEqual(len(ips), 1)
        self.assertEqual(ips[0].ip_address, '185.102.95.236')

    def test_ex_rename_node(self):
        node = self.driver.list_nodes()[0]
        self.assertTrue(self.driver.ex_rename_node(node, name='new-name'))

    def test_ex_rename_volume(self):
        volume = self.driver.list_volumes()[0]
        self.assertTrue(self.driver.ex_rename_volume(volume, name='new-name'))

    def test_ex_network(self):
        network = self.driver.ex_list_networks()[0]
        self.assertTrue(self.driver.ex_rename_network(network, name='new-name'))


class GridscaleMockHttp(MockHttp):
    fixtures = ComputeFileFixtures('gridscale')

    def _objects_servers(self, method, url, body, headers):
        body = self.fixtures.load('list_nodes.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_locations(self, method, url, body, headers):
        body = self.fixtures.load('list_locations.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_storages(self, method, url, body, headers):
        body = self.fixtures.load('list_volumes.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_storages_DELETE(self, method, url, body, headers):
        # test_destroy_node_success
        body = self.fixtures.load('list_volumes.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_ips_DELETE(self, method, url, body, headers):
        # test_destroy_node_success
        body = self.fixtures.load('ex_list_ips.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_ips_56b8d161_325b_4fd4_DELETE(self, method, url, body, headers):
        # test_destroy_node_success
        return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])

    def _objects_storages_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load('list_volumes_empty.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_templates(self, method, url, body, headers):
        body = self.fixtures.load('list_images.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_sshkeys(self, method, url, body, headers):
        body = self.fixtures.load('list_key_pairs.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_snapshots(
        self, method, url, body, headers):
        body = self.fixtures.load('list_volume_snapshots.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f(
        self, method, url, body, headers):
        # test_ex_rename_node
        if method == 'PATCH':
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee(
        self, method, url, body, headers):
        # test_ex_rename_node
        if method == 'PATCH':
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_networks_1196529b_a8de_417f(
        self, method, url, body, headers):
        # test_ex_rename_network
        if method == 'PATCH':
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_servers_POST(self, method, url, body, headers):
        # create_node
        if method == 'POST':
            body = self.fixtures.load('create_node.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _requests_x123xx1x_123x_1x12_123x_123xxx123x1x_POST(self, method, url,
                                                            body, headers):
        body = self.fixtures.load('create_node_response_dict.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_690de890_13c0_4e76_8a01_e10ba8786e53_POST(self, method, url,
                                                            body, headers):
        body = self.fixtures.load('create_node_dict.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # create_volume
    def _objects_storages_POST(self, method, url, body, headers):
        body = self.fixtures.load('create_volume.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_storages_690de890_13c0_4e76_8a01_e10ba8786e53_POST(self, method,
                                                                    url, body, headers):
        body = self.fixtures.load('create_volume_response_dict.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_ips_POST(self, method, url, body, headers):
        # ex_create_ip
        if method == 'POST':
            body = self.fixtures.load('create_ip.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_ips_690de890_13c0_4e76_8a01_e10ba8786e53_POST(self, method, url,
                                                                body, headers):
        body = self.fixtures.load('create_ip_response_dict.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_storages_POST(self, method, url, body, headers):
        body = self.fixtures.load('volume_to_node.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_ips_POST(self, method, url, body, headers):
        body = self.fixtures.load('ips_to_node.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_networks_POST(self, method, url, body, headers):
        body = self.fixtures.load('create_network.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_networks_1196529b_a8de_417f_DELETE(self, method, url, body, headers):
        # test_ex_destroy_network
        if method == 'DELETE':
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_networks_POST(self, method, url, body, headers):
        body = self.fixtures.load('network_to_node.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_power_POST(self, method, url, body, headers):
        if method == 'PATCH':
            body = self.fixtures.load('ex_start_node.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_POST(self, method, url, body, headers):
        body = self.fixtures.load('create_node_dict.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_networks(self, method, url, body, headers):
        if method == 'GET':
            body = self.fixtures.load('ex_list_networks.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_ips(self, method, url, body, headers):
        if method == 'GET':
            body = self.fixtures.load('ex_list_ips.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_servers_1479405e_d46c_47a2_91e8_eb43951c899f_DELETE(self, method, url, body, headers):
        if method == 'DELETE':
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_DELETE(self, method, url, body, headers):
        if method == 'DELETE':
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_snapshots_d755de62_4d75_4d61_addd_a5c9743a5deb_DELETE(self, method, url, body, headers):
        if method == 'DELETE':
            return (httplib.NO_CONTENT, None, {}, httplib.responses[httplib.NO_CONTENT])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_templates_POST(self, method, url, body, headers):
        # create_image
        if method == 'POST':
            body = self.fixtures.load('create_image.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_snapshots_POST(self, method, url, body, headers):
        # create_image
        if method == 'POST':
            body = self.fixtures.load('create_image.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_storages_e66bb753_4a03_4ee2_a069_a601f393c9ee_snapshots_690de890_13c0_4e76_8a01_e10ba8786e53_POST(self, method, url, body, headers):
        # create_image
        body = self.fixtures.load('create_image_dict.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _objects_templates_690de890_13c0_4e76_8a01_e10ba8786e53_POST(self, method, url, body, headers):
        # create_image
        if method == 'GET':
            body = self.fixtures.load('create_image_dict.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError('Invalid method')

    def _objects_templates_12345(self, method, url, body, headers):
        # get_image
        if method == 'GET':
            body = self.fixtures.load('get_image.json')
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:  # pragma: no cover
            raise ValueError('Invalid method')


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A driver for cloudscale.ch.
"""

import json

from libcloud.utils.py3 import httplib

from libcloud.common.base import ConnectionKey, JsonResponse
from libcloud.compute.types import Provider, NodeState
from libcloud.common.types import InvalidCredsError
from libcloud.compute.base import NodeDriver
from libcloud.compute.base import Node, NodeImage, NodeSize


class CloudscaleResponse(JsonResponse):
    valid_response_codes = [httplib.OK, httplib.ACCEPTED, httplib.CREATED,
                            httplib.NO_CONTENT]

    def parse_error(self):
        body = self.parse_body()
        if self.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError(body['detail'])
        else:
            # We are taking the first issue here. There might be multiple ones,
            # but that doesn't really matter. It's nicer if the error is just
            # one error (because it's a Python API and there's only one
            # exception.
            return next(iter(body.values()))

    def success(self):
        return self.status in self.valid_response_codes


class CloudscaleConnection(ConnectionKey):
    """
    Connection class for the cloudscale.ch driver.
    """
    host = 'api.cloudscale.ch'
    responseCls = CloudscaleResponse

    def add_default_headers(self, headers):
        """
        Add headers that are necessary for every request

        This method adds ``token`` to the request.
        """
        headers['Authorization'] = 'Bearer %s' % (self.key)
        headers['Content-Type'] = 'application/json'
        return headers


class CloudscaleNodeDriver(NodeDriver):
    """
    Cloudscale's node driver.
    """

    connectionCls = CloudscaleConnection

    type = Provider.CLOUDSCALE
    name = 'Cloudscale'
    website = 'https://www.cloudscale.ch'

    NODE_STATE_MAP = dict(
        changing=NodeState.PENDING,
        running=NodeState.RUNNING,
        stopped=NodeState.STOPPED,
        paused=NodeState.PAUSED,
    )

    def __init__(self, key, **kwargs):
        super(CloudscaleNodeDriver, self).__init__(key, **kwargs)

    def list_nodes(self):
        """
        List all your existing compute nodes.
        """
        return self._list_resources('/v1/servers', self._to_node)

    def list_sizes(self):
        """
        Lists all available sizes. On cloudscale these are known as flavors.
        """
        return self._list_resources('/v1/flavors', self._to_size)

    def list_images(self):
        """
        List all images.

        Images are identified by slugs on cloudscale.ch. This means that minor
        version upgrades (e.g. Ubuntu 16.04.1 to Ubuntu 16.04.2) will be
        possible within the same id ``ubuntu-16.04``.
        """
        return self._list_resources('/v1/images', self._to_image)

    def create_node(self, name, size, image, location=None,
                    ex_create_attr=None):
        """
        Create a node.

        The `ex_create_attr` parameter can include the following dictionary
        key and value pairs:

        * `ssh_keys`: ``list`` of ``str`` ssh public keys
        * `volume_size_gb`: ``int`` defaults to 10.
        * `bulk_volume_size_gb`: defaults to None.
        * `use_public_network`: ``bool`` defaults to True
        * `use_private_network`: ``bool`` defaults to False
        * `use_ipv6`: ``bool`` defaults to True
        * `anti_affinity_with`: ``uuid`` of a server to create an anti-affinity
          group with that server or add it to the same group as that server.
        * `user_data`: ``str`` for optional cloud-config data

        :keyword ex_create_attr: A dictionary of optional attributes for
                                 droplet creation
        :type ex_create_attr: ``dict``

        :return: The newly created node.
        :rtype: :class:`Node`
        """
        ex_create_attr = ex_create_attr or {}
        attr = dict(ex_create_attr)
        attr.update(
            name=name,
            image=image.id,
            flavor=size.id,
        )
        result = self.connection.request(
            '/v1/servers',
            data=json.dumps(attr),
            method='POST'
        )
        return self._to_node(result.object)

    def reboot_node(self, node):
        """
        Reboot a node. It's also possible to use ``node.reboot()``.
        """
        return self._action(node, 'reboot')

    def start_node(self, node):
        """
        Start a node. This is only possible if the node is stopped.
        """
        return self._action(node, 'start')

    def stop_node(self, node):
        """
        Stop a specific node. Similar to ``shutdown -h now``. This is only
        possible if the node is running.
        """
        return self._action(node, 'stop')

    def ex_start_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0
        return self.start_node(node=node)

    def ex_stop_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0
        return self.stop_node(node=node)

    def ex_node_by_uuid(self, uuid):
        """
        :param str ex_user_data: A valid uuid that references your exisiting
            cloudscale.ch server.
        :type       ex_user_data:  ``str``

        :return: The server node you asked for.
        :rtype: :class:`Node`
        """
        res = self.connection.request(self._get_server_url(uuid))
        return self._to_node(res.object)

    def destroy_node(self, node):
        """
        Delete a node. It's also possible to use ``node.destroy()``.
        This will irreversibly delete the cloudscale.ch server and all its
        volumes. So please be cautious.
        """
        res = self.connection.request(
            self._get_server_url(node.id),
            method='DELETE'
        )
        return res.status == httplib.NO_CONTENT

    def _get_server_url(self, uuid):
        return '/v1/servers/%s' % uuid

    def _action(self, node, action_name):
        response = self.connection.request(
            self._get_server_url(node.id) + '/' + action_name,
            method='POST'
        )
        return response.status == httplib.OK

    def _list_resources(self, url, tranform_func):
        data = self.connection.request(url, method='GET').object
        return [tranform_func(obj) for obj in data]

    def _to_node(self, data):
        state = self.NODE_STATE_MAP.get(data['status'], NodeState.UNKNOWN)
        extra_keys_exclude = ['uuid', 'name', 'status', 'flavor', 'image']
        extra = {}
        for k, v in data.items():
            if k not in extra_keys_exclude:
                extra[k] = v

        public_ips = []
        private_ips = []
        for interface in data['interfaces']:
            if interface['type'] == 'public':
                ips = public_ips
            else:
                ips = private_ips
            for address_obj in interface['addresses']:
                ips.append(address_obj['address'])

        return Node(
            id=data['uuid'],
            name=data['name'],
            state=state,
            public_ips=public_ips,
            private_ips=private_ips,
            extra=extra,
            driver=self,
            image=self._to_image(data['image']),
            size=self._to_size(data['flavor']),
        )

    def _to_size(self, data):
        extra = {'vcpu_count': data['vcpu_count']}
        ram = data['memory_gb'] * 1024

        return NodeSize(id=data['slug'], name=data['name'],
                        ram=ram, disk=10,
                        bandwidth=0, price=0,
                        extra=extra, driver=self)

    def _to_image(self, data):
        extra = {'operating_system': data['operating_system']}
        return NodeImage(id=data['slug'], name=data['name'], extra=extra,
                         driver=self)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
1&1 Cloud Server Compute driver
"""
import json

from libcloud.compute.providers import Provider
from libcloud.common.base import JsonResponse, ConnectionKey
from libcloud.compute.base import NodeSize, NodeImage, NodeLocation, \
    Node, NodeAuthPassword, NodeAuthSSHKey
from libcloud.common.types import InvalidCredsError
from libcloud.compute.types import NodeState
from libcloud.utils.py3 import httplib
from libcloud.compute.base import NodeDriver

from time import sleep

API_HOST = 'cloudpanel-api.1and1.com'
API_VERSION = '/v1/'

__all__ = [
    'API_HOST',
    'API_VERSION',
    'OneAndOneResponse',
    'OneAndOneConnection',
    'OneAndOneNodeDriver'
]


class OneAndOneResponse(JsonResponse):
    """
    OneAndOne response parsing.
    """
    valid_response_codes = [httplib.OK, httplib.CREATED, httplib.ACCEPTED]

    def parse_error(self):

        if self.status == httplib.UNAUTHORIZED:
            body = self.parse_body()
            raise InvalidCredsError(body['message'])
        else:
            body = self.parse_body()
            if 'message' in body:
                error = '%s (code: %s)' % (body['message'], self.status)
            else:
                error = body
            return error

    def success(self):
        return self.status in self.valid_response_codes


class OneAndOneConnection(ConnectionKey):
    """
    Connection class for the 1&1 driver
    """

    host = API_HOST
    api_prefix = API_VERSION
    responseCls = OneAndOneResponse

    def encode_data(self, data):
        return json.dumps(data)

    def add_default_headers(self, headers):
        """
        Add headers that are necessary for every request

        This method adds ``token`` and ``Content-Type`` to the request.
        """
        headers['X-Token'] = self.key
        headers['Content-Type'] = 'application/json'
        return headers

    def request(self, action, params=None, data=None, headers=None,
                method='GET', raw=False):
        """
        Some requests will use the href attribute directly.
        If this is not the case, then we should formulate the
        url based on the action specified.
        If we are using a full url, we need to remove the
        host and protocol components.
        """
        action = self.api_prefix + action.lstrip('/')

        return super(OneAndOneConnection, self). \
            request(action=action,
                    params=params,
                    data=data,
                    headers=headers,
                    method=method,
                    raw=raw)


class OneAndOneNodeDriver(NodeDriver):
    """
      Base OneAndOne node driver.
    """
    connectionCls = OneAndOneConnection
    name = '1and1'
    website = 'http://www.1and1.com'
    type = Provider.ONEANDONE

    NODE_STATE_MAP = {
        'POWERING_ON': NodeState.STARTING,
        'POWERING_OFF': NodeState.PENDING,
        'POWERED_OFF': NodeState.STOPPING,
        'POWERED_ON': NodeState.RUNNING,
        'REBOOTING': NodeState.REBOOTING,
        'CONFIGURING': NodeState.RECONFIGURING,
        'REMOVING': NodeState.UNKNOWN,
        'DEPLOYING': NodeState.STARTING,
    }

    """
    Core Functions
    """

    def list_sizes(self):
        """
        Lists all sizes

        :return: A list of all configurable node sizes.
        :rtype: ``list`` of :class:`NodeSize`
        """
        sizes = []

        fixed_instances = self._list_fixed_instances()
        for value in fixed_instances:
            node_size = self._to_node_size(value)
            sizes.append(node_size)

        return sizes

    def list_locations(self):
        """
        Lists all locations

        :return: ``list`` of :class:`NodeLocation`
        :rtype: ``list``
        """
        datacenters = self.ex_list_datacenters()
        locations = []
        for values in datacenters:
            node_size = self._to_location(values)
            locations.append(node_size)

        return locations

    def list_images(self, image_type=None):
        """
        :return: ``list`` of :class: `NodeImage`
        :rtype: ``list``
        """
        response = self.connection.request(
            action='server_appliances',
            method='GET'
        )

        return self._to_images(response.object, image_type)

    def get_image(self, image_id):
        response = self.connection.request(
            action='server_appliances/%s' % image_id,
            method='GET'
        )
        return self._to_image(response.object)

    """
    Node functions
    """

    def create_node(self,
                    name,
                    image,
                    ex_fixed_instance_size_id,
                    location=None,
                    auth=None,
                    ex_ip=None,
                    ex_monitoring_policy_id=None,
                    ex_firewall_policy_id=None,
                    ex_loadbalancer_id=None,
                    ex_description=None,
                    ex_power_on=None):
        """
        Creates a node.

        :param name: The name of the new node
        :type name: `str`

        :param ex_fixed_instance_size_id:
        Fixed instance size ID from list_sizes
        :type ex_fixed_instance_size_id: ``str``

        :param location: 1&1 Data center Location
        :type location: `NodeLocation`

        :param ex_ip: IP address
        :type ex_ip: `str`

        :param ex_ssh_key: SSH Key
        :type ex_ssh_key: `str`

        :param password: Password
        :type password: `str`

        :param ex_monitoring_policy_id:
        :type ex_firewall_policy_id: `str`

        :param ex_firewall_policy_id:
        :type ex_firewall_policy_id: `str`

        :param ex_loadbalancer_id:
        :type ex_loadbalancer_id: `str`

        :param ex_description:
        :type ex_description: `str`

        :param ex_power_on:
        :type ex_power_on: `bool`

        :return:    Instance of class ``Node``
        :rtype:     :class:`Node`
        """

        body = {
            'name': name,
            'appliance_id': image.id,
            'hardware': {
                'fixed_instance_size_id': ex_fixed_instance_size_id
            },
        }

        if location is not None:
            body['datacenter_id'] = location.id
        if ex_power_on is not None:
            body['power_on'] = ex_power_on

        if ex_description is not None:
            body['description'] = ex_description

        if ex_firewall_policy_id is not None:
            body['firewall_policy_id'] = ex_firewall_policy_id

        if ex_monitoring_policy_id is not None:
            body['monitoring_policy_id'] = ex_monitoring_policy_id

        if ex_loadbalancer_id is not None:
            body['loadbalancer_id'] = ex_loadbalancer_id

        if auth is not None:
            if isinstance(auth, NodeAuthPassword):
                body['password'] = auth.password
            elif isinstance(auth, NodeAuthSSHKey):
                body['rsa_key'] = auth.pubkey
        if ex_ip is not None:
            body['ip_id'] = ex_ip

        response = self.connection.request(
            action='servers',
            data=body,
            method='POST',
        )

        return self._to_node(response.object)

    def list_nodes(self):
        """
        List all nodes.

        :return: ``list`` of :class:`Node`
        :rtype: ``list``
        """
        response = self.connection.request(
            action='servers',
            method='GET'
        )

        return self._to_nodes(response.object)

    def destroy_node(self, node, ex_keep_ips=False):
        """
        Destroys a node.

        :param node: The node you wish to destroy.
        :type volume: :class:`Node`

        :param ex_keep_ips: True to keep all IP addresses assigned to the node
        :type ex_keep_ips: : ``bool``

        :return:    Instance of class ``Node``
        :rtype:     :class: `Node`
        """
        self.ex_shutdown_server(node.id)

        self._wait_for_state(node.id, 'POWERED_OFF')

        response = self.connection.request(
            action='servers/%s' % node.id,
            params={'keep_ips': ex_keep_ips},
            method='DELETE'
        )

        return self._to_node(response.object)

    def reboot_node(self, node):
        """
        Reboots the node.

        :param node: The node you wish to destroy.
        :type volume: :class:`Node`

        :return:    Instance of class ``Node``
        :rtype:     :class: `Node`
        """
        shutdown_body = {
            "action": "REBOOT",
            "method": "HARDWARE"
        }
        response = self.connection.request(
            action='servers/%s/status/action' % node.id,
            data=shutdown_body,
            method='PUT',
        )
        return self._to_node(response.object)

    """
    Extension functions
    """

    def ex_rename_server(self, server_id, name=None, description=None):
        """
        Renames the server
        :param  server_id: ID of the server you want to rename

        :param  name: New name of the server
        :type: ``str``

        :param description: New description of the server
        :type: ``str``

        :return:    Instance of class ``Node``
        :rtype:     :class: `Node`
        """

        body = {}
        if name is not None:
            body["name"] = name
        if description is not None:
            body["description"] = description

        response = self.connection.request(
            action='servers/%s' % server_id,
            data=body,
            method='PUT'
        )

        return self._to_node(response.object)

    def ex_get_server_hardware(self, server_id):
        """
        Gets all server hardware

        :param server_id: Id of the server
        :type: ``str``

        :return: Server's hardware
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='servers/%s/hardware' % server_id,
            method='GET'
        )
        return response.object

    """
    Hardware operations
    """

    def ex_modify_server_hardware(self, server_id,
                                  fixed_instance_size_id=None, vcore=None,
                                  cores_per_processor=None, ram=None):
        """
        Modifies server's hardware

        :param server_id:
        :type: ``str``

        :param fixed_instance_size_id: Id of the fixed instance size
        :type: ``str``

        :param vcore: Virtual cores count
        :type: ``int``

        :param cores_per_processor: Count of cores per procesor
        :type: ``int``

        :param ram: Amount of ram for the server
        :type: ``int``

        :return:    Instance of class ``Node``
        :type:     :class: `Node`
        """

        body = {}

        if fixed_instance_size_id is not None:
            body['fixed_instance_size_id'] = fixed_instance_size_id
        if vcore is not None:
            body['vcore'] = vcore
        if cores_per_processor is not None:
            body['cores_per_processor'] = cores_per_processor
        if ram is not None:
            body['ram'] = ram

        response = self.connection.request(
            action='servers/%s/hardware' % server_id,
            data=body,
            method='PUT'
        )

        return self._to_node(response.object)

    """
    HDD operations
    """

    def ex_modify_server_hdd(self, server_id, hdd_id=None, size=None):
        """
        Modifies server hard disk drives

        :param server_id: Id of the server
        :type: ``str``

        :param hdd_id: Id of the hard disk
        :type: ``str``

        :param size: Size of the hard disk
        :type: ``str``

        :return:    Instance of class ``Node``
        :rtype:     :class: `Node`
        """

        body = {}

        if size is not None:
            body['size'] = size

            response = self.connection.request(
                action='servers/%s/hardware/hdds/%s' % (server_id, hdd_id),
                data=body,
                method='PUT'
            )

            return self._to_node(response.object)

    def ex_add_hdd(self, server_id, size, is_main):
        """
        Add a hard disk to the server

        :param server_id: Id of the server
        :type: ``str``

        :param size: Size of the new disk
        :type: ``str``

        :param is_main: Indicates if the disk is going to be the boot disk
        :type: ``boolean``

        :return:    Instance of class ``Node``
        :type:     :class: `Node`
        """
        body = {
            'size': size,
            'is_main': is_main
        }

        response = self.connection.request(
            action='servers/%s/hardware/hdds' % server_id,
            data=body,
            method='POST'
        )

        return self._to_node(response.object)

    def ex_remove_hdd(self, server_id, hdd_id):
        """
        Removes existing hard disk

        :param server_id: Id of the server
        :type: ``str``

        :param hdd_id: Id of the hard disk
        :type: ``str``

        :return:    Instance of class ``Node``
        :rtype:     :class: `Node`
        """

        response = self.connection.request(
            action='servers/%s/hardware/hdds/%s' % (server_id, hdd_id),
            method='DELETE'
        )

        return self._to_node(response.object)

    """
    Data center operations
    """

    def ex_list_datacenters(self):
        """
        Lists all data centers

        :return: List of data centers
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='datacenters',
            method='GET'
        )

        return response.object

    def ex_get_server(self, server_id):
        """
        Gets a server

        :param server_id: Id of the server to be retrieved
        :type: ``str``

        :return:    Instance of class ``Node``
        :rtype:     :class: `Node`
        """

        response = self.connection.request(
            action='servers/%s' % (server_id),
            method='GET'
        )

        return self._to_node(response.object)

    def ex_shutdown_server(self, server_id, method='SOFTWARE'):
        """
        Shuts down the server

        :param server_id: Id of the server to be shut down
        :type: ``str``

        :param method: Method of shutting down "SOFTWARE" or "HARDWARE"

        :return:    Instance of class ``Node``
        :rtype:     :class: `Node`
        """

        shutdown_body = {
            'action': 'POWER_OFF',
            'method': method
        }
        response = self.connection.request(
            action='servers/%s/status/action' % (server_id),
            data=shutdown_body,
            method='PUT',
        )
        return self._to_node(response.object)

    """
    Image operations
    """

    def ex_get_server_image(self, server_id):
        """
        Gets server image

        :param server_id: Id of the server
        :type: ``str``

        :return: Server image
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='servers/%s/image' % server_id,
            method='GET'
        )
        return response.object

    def ex_reinstall_server_image(self, server_id, image_id, password=None):
        """
        Installs a new image on the server

        :param server_id: Id of the server
        :type: ``str``

        :param image_id: Id of the image (Server Appliance)
        :type: ``str``

        :param password: New password for the server

        :return:    Instance of class ``Node``
        :rtype:     :class: `Node`
        """

        body = {
            'id': image_id,
        }

        if password is not None:
            body['password'] = password

        response = self.connection.request(
            action='servers/%s/image' % server_id,
            data=body,
            method='PUT'
        )
        return self._to_node(response.object)

    """
    Server IP operations
    """

    def ex_list_server_ips(self, server_id):
        """
        Gets all server IP objects

        :param server_id: Id of the server
        :type: ``str``

        :return: List of server IP objects
        :rtype: ``list`` of ``dict``
        """
        response = self.connection.request(
            action='servers/%s/ips' % server_id,
            method='GET'
        )

        return response.object

    def ex_get_server_ip(self, server_id, ip_id):
        """
        Get a single server IP object

        :param server_id: Id of the server
        :type: ``str``

        :param ip_id: ID of the IP address
        :type: ``str``

        :return: IP address object
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='servers/%s/ips/%s' % (server_id, ip_id),
            method='GET'
        )

        return response.object

    def ex_assign_server_ip(self, server_id, ip_type):
        """
        Assigns a new IP address to the server

        :param server_id: Id of the server
        :type: ``str``

        :param ip_type: Type of the IP address [IPV4,IPV6]
        :type: ``str``

        :return: ``Node`` instance
        :rtype: ``Node``
        """

        body = {
            'type': ip_type
        }

        response = self.connection.request(
            action='servers/%s/ips' % server_id,
            data=body,
            method='POST'
        )

        return self._to_node(response.object)

    def ex_remove_server_ip(self, server_id, ip_id, keep_ip=None):
        """
        Removes an IP address from the server

        :param server_id: Id of the server
        :type: ``str``

        :param ip_id: ID of the IP address
        :type: ``str``

        :param keep_ip: Indicates whether IP address will be removed from
                        the Cloud Panel
        :type: ``boolean``

        :return: ``Node`` instance
        :rtype: ``Node``
        """

        body = {}
        if keep_ip is not None:
            body['keep_ip'] = keep_ip

        response = self.connection.request(
            action='servers/%s/ips/%s' % (server_id, ip_id),
            data=body,
            method='DELETE'
        )

        return self._to_node(response.object)

    def ex_get_server_firewall_policies(self, server_id, ip_id):
        """
        Gets a firewall policy of attached to the server's IP

        :param server_id: Id of the server
        :type: ``str``

        :param ip_id: ID of the IP address
        :type: ``str``

        :return: IP address object
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='/servers/%s/ips/%s/firewall_policy' % (server_id, ip_id),
            method='GET'
        )

        return response.object

    def ex_add_server_firewall_policy(self, server_id, ip_id, firewall_id):
        """
        Adds a firewall policy to the server's IP address

        :param server_id: Id of the server
        :type: ``str``

        :param ip_id: ID of the IP address
        :type: ``str``

        :param firewall_id: ID of the firewall policy
        :type: ``str``

        :return: ``Node`` instance
        :rtype: ``Node``
        """
        body = {
            'id': firewall_id
        }
        response = self.connection.request(
            action='/servers/%s/ips/%s/firewall_policy' % (server_id, ip_id),
            data=body,
            method='POST'
        )

        return self._to_node(response.object)

    """
    Firewall Policy operations
    """

    def ex_create_firewall_policy(self, name, rules, description=None):
        """
        Creates a firewall Policy.

        :param name:
        :param description:
        :param rules:

        :rtype: `dict`
        :return: `dict` firewall policy

        """
        body = {
            'name': name
        }

        if description is not None:
            body['description'] = description

        if len(rules) == 0:
            raise ValueError(
                'At least one firewall rule is required.'
            )
        else:
            body['rules'] = rules

        response = self.connection.request(
            action='firewall_policies',
            data=body,
            method='POST',
        )

        return response.object

    def ex_list_firewall_policies(self):
        """"
        List firewall policies

        :return: 'dict'
        """

        response = self.connection.request(
            action='firewall_policies',
            method='GET'
        )

        return response.object

    def ex_get_firewall_policy(self, fw_id):
        """
        Gets firewall policy

        :param fw_id: ID of the firewall policy
        :return: 'dict'
        """

        response = self.connection.request(
            action='firewall_policy/%s' % fw_id,
            method='GET'
        )

        return response.object

    def ex_delete_firewall_policy(self, fw_id):
        """
        Deletes firewall policy

        :param fw_id: ID of the Firewall
        :return: 'dict'
        """
        response = self.connection.request(
            action='firewall_policy/%s' % fw_id,
            method='DELETE'
        )

        return response.object

    """
    Shared storage operations
    """

    def ex_list_shared_storages(self):
        """
        List of shared storages
        :return: 'dict'
        """
        response = self.connection.request(
            action='shared_storages',
            method='GET'
        )

        return response.object

    def ex_get_shared_storage(self, storage_id):
        """
        Gets a shared storage
        :return: 'dict'
        """
        response = self.connection.request(
            action='shared_storages/%s' % (storage_id),
            method='GET'
        )

        return response.object

    def ex_create_shared_storage(self, name, size, datacenter_id=None,
                                 description=None):
        """
        Creates a shared storage
        :param name: Name of the storage
        :param size: Size of the storage
        :param datacenter_id: datacenter where storage should be created
        :param description: description ot the  storage
        :return: 'dict'
        """

        body = {
            'name': name,
            'size': size,
            'datacenter_id': datacenter_id
        }

        if description is not None:
            body['description'] = description

        response = self.connection.request(
            action='shared_storages',
            data=body,
            method='POST'
        )

        return response.object

    def ex_delete_shared_storage(self, storage_id):
        """
        Removes a shared storage

        :param storage_id: Id of the shared storage
        :type: ``str``

        :return: Instnace of shared storage
        :rtype: ``list`` of ``dict``
        """
        response = self.connection.request(
            action='shared_storages/%s' % storage_id,
            method='DELETE'
        )

        return response.object

    def ex_attach_server_to_shared_storage(self, storage_id,
                                           server_id, rights):
        """
        Attaches a single server to a shared storage

        :param storage_id: Id of the shared storage
        :param server_id: Id of the server to be attached to the shared storage
        :param rights:
        :return:
        :rtype: 'dict'
        """
        body = {
            'severs': [
                {
                    'id': server_id,
                    'rights': rights
                }
            ]
        }

        response = self.connection.request(
            action='shared_storages/%s/servers' % storage_id,
            data=body,
            method='POST'
        )

        return response.object

    def ex_get_shared_storage_server(self, storage_id, server_id):
        """
        Gets a shared storage's server
        :param storage_id:
        :param server_id:
        :return:
        """
        response = self.connection.request(
            action='shared_storages/%s/servers/%s' % (storage_id, server_id),
        )

        return response.object

    def ex_detach_server_from_shared_storage(self, storage_id,
                                             server_id):
        """
        Detaches a server from shared storage

        :param storage_id: Id of the shared storage
        :type: ``str``

        :param server_id: Id of the server
        :type: ``str``

        :return: Instance of shared storage
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='shared_storages/%s/servers/%s' % (storage_id, server_id),
            method='DELETE'
        )

        return response.object

    """
    Load Balancers operations
    """

    def ex_create_load_balancer(self, name, method, rules,
                                persistence=None,
                                persistence_time=None,
                                health_check_test=None,
                                health_check_interval=None,
                                health_check_path=None,
                                health_check_parser=None,
                                datacenter_id=None,
                                description=None):
        """

        :param name: Name of the load balancer

        :param method: Load balancer method

        :param rules: Load balancer rules
        :type rules: ``list`` of ``dict``

        :param persistence: Indictes if persistance is set
        :type persistence: ``boolean``

        :param persistence_time: Persistance time
        :type persistence_time: ``int``

        :param health_check_test: Type of test
        :type health_check_test:``str``

        :param health_check_interval: Interval of the check

        :param health_check_path: Path
        :type health_check_path: ``str``

        :param health_check_parser: Parser
        :type health_check_parser:``str``

        :param datacenter_id: Data center id
        :type datacenter_id:``str``

        :param description: Description of load balancer
        :type description:``str``

        :return: ``dict``
        """

        body = {
            'name': name,
            'method': method,
        }

        body['rules'] = []
        body['rules'] = rules

        if persistence is not None:
            body['persistence'] = persistence
        if persistence_time is not None:
            body['persistence_time'] = persistence_time
        if health_check_test is not None:
            body['health_check_test'] = health_check_test
        if health_check_interval is not None:
            body['health_check_interval'] = health_check_interval
        if health_check_path is not None:
            body['health_check_path'] = health_check_path
        if health_check_parser is not None:
            body['health_check_parser'] = health_check_parser
        if datacenter_id is not None:
            body['datacenter_id'] = datacenter_id
        if description is not None:
            body['description'] = description

        response = self.connection.request(
            action='load_balancers',
            data=body,
            method='POST'
        )

        return response.object

    def ex_update_load_balancer(self, lb_id, name=None, description=None,
                                health_check_test=None,
                                health_check_interval=None,
                                persistence=None,
                                persistence_time=None,
                                method=None):
        body = {}

        if name is not None:
            body['name'] = name
        if description is not None:
            body['description'] = description
        if health_check_test is not None:
            body['health_check_test'] = health_check_test
        if health_check_interval is not None:
            body['health_check_interval'] = health_check_interval
        if persistence is not None:
            body['persistence'] = persistence
        if persistence_time is not None:
            body['persistence_time'] = persistence_time
        if method is not None:
            body['method'] = method

        response = self.connection.request(
            action='load_balancers/%s' % lb_id,
            data=body,
            method='PUT'
        )

        return response.object

    def ex_add_servers_to_load_balancer(self, lb_id, server_ips=[]):
        """
        Adds server's IP address to load balancer

        :param lb_id: Load balancer ID
        :type: ``str``

        :param server_ips: Array of server IP IDs
        :type: ``list`` of ``str``

        :return: Instance of load balancer
        :rtype: ``dict``
        """
        body = {
            'server_ips': server_ips,
        }

        response = self.connection.request(
            action='load_balancers/%s/server_ips' % lb_id,
            data=body,
            method='POST'
        )

        return response.object

    def ex_remove_server_from_load_balancer(self, lb_id, server_ip):
        """
        Removes server's IP from load balancer

        :param lb_id: Load balancer ID
        :type: ``str``

        :param server_ip: ID of the server IP
        :type: ``str``

        :return: Instance of load balancer
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='/load_balancers/%s/server_ips/%s' % (lb_id, server_ip),
            method='DELETE'
        )

        return response.object

    def ex_add_load_balancer_rule(self, lb_id, protocol, port_balancer,
                                  port_server, source=None):
        """
        Adds a rule to load balancer

        :param lb_id: Load balancer ID
        :rtype: ``str``

        :param protocol: Load balancer protocol
        :rtype: ``str``

        :param port_balancer: Port to be balananced
        :rtype: ``int``

        :param port_server: Server port
        :rtype: ``int``

        :param source: Source IP address
        :rtype: ``str``

        :return: Instance of load balancer
        :rtype: ``dict``
        """

        body = {
            'rules': [
                {
                    'protocol': protocol,
                    'port_balancer': port_balancer,
                    'port_server': port_server
                }
            ]
        }

        if source is not None:
            body['rules'][0]['source'] = source

        response = self.connection.request(
            action='/load_balancers/%s/rules' % lb_id,
            data=body,
            method='POST'
        )

        return response.object

    def ex_remove_load_balancer_rule(self, lb_id, rule_id):
        """
        Removes load balancer rule

        :param lb_id: Load balancer ID
        :rtype: ``str``

        :param rule_id: Rule ID
        :rtype: ``str``

        :return: Instance of load balancer
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='/load_balancers/%s/rules/%s' % (lb_id, rule_id),
            method='DELETE'
        )

        return response.object

    def ex_list_load_balancers(self):
        """
        Lists all load balancers

        :return: List of load balancers
        :rtype: ``list`` of ``dict``
        """
        response = self.connection.request(
            action='load_balancers',
            method='GET'
        )
        return response.object

    def ex_get_load_balancer(self, lb_id):
        """
        Gets a single load balancer

        :param lb_id: ID of the load balancer
        :type lb_id: ``str``

        :return: Instance of load balancer
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='load_balancers/%s' % lb_id,
            method='GET'
        )

        return response.object

    def ex_list_load_balancer_server_ips(self, lb_id):
        """
        List balanced server IP addresses

        :param lb_id: ID of the load balancer
        :type lb_id: ``str``

        :return: Array of IP address IDs
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='load_balancers/%s/server_ips' % lb_id,
            method='GET'
        )

        return response.object

    def ex_get_load_balancer_server_ip(self, lb_id, server_ip):
        """
        Gets load balanced server id

        :param lb_id: ID of the load balancer
        :type lb_id: ``str``

        :param server_ip: ID of the server IP
        :type server_ip: ``str``

        :return: Server IP
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='load_balancers/%s/server_ips/%s' % (lb_id, server_ip),
            method='GET'
        )

        return response.object

    def ex_list_load_balancer_rules(self, lb_id):
        """
        Lists loadbalancer rules

        :param lb_id: ID of the load balancer
        :type lb_id: ``str``

        :return: Lists of rules
        :rtype: ``list`` of ``dict``
        """
        response = self.connection.request(
            action='load_balancers/%s/rules' % lb_id,
            method='GET'
        )

        return response.object

    def ex_get_load_balancer_rule(self, lb_id, rule_id):
        """
        Get a load balancer rule

        :param lb_id: ID of the load balancer
        :type lb_id: ``str``

        :param rule_id: Rule ID
        :type rule_id: ``str``

        :return: A load balancer rule
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='load_balancers/%s/rules/%s' % (lb_id, rule_id),
            method='GET'
        )

        return response.object

    def ex_delete_load_balancer(self, lb_id):
        """
        Deletes a load balancer rule

        :param lb_id: ID of the load balancer
        :type lb_id: ``str``

        :param rule_id: Rule ID
        :type rule_id: ``str``

        :return: Instance of load balancer
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='load_balancers/%s' % lb_id,
            method='DELETE'
        )

        return response.object

    """
    Public IP operations
    """

    def ex_list_public_ips(self):
        """
        Lists all public IP addresses

        :return: Array of public addresses
        :rtype: ``list`` of ``dict``
        """
        response = self.connection.request(
            action='public_ips',
            method='GET'
        )

        return response.object

    def ex_create_public_ip(self, type, reverse_dns=None, datacenter_id=None):
        """
        Creates a public IP

        :param type: Type of IP (IPV4 or IPV6)
        :type type: ``str``

        :param reverse_dns: Reverse DNS
        :type reverse_dns: ``str``

        :param datacenter_id: Datacenter ID where IP address will be crated
        :type datacenter_id: ``str``

        :return: Instance of Public IP
        :rtype: ``dict``
        """
        body = {
            'type': type
        }

        if reverse_dns is not None:
            body['reverse_dns'] = reverse_dns
        if datacenter_id is not None:
            body['datacenter_id'] = datacenter_id

        response = self.connection.request(
            action='public_ips',
            data=body,
            method='POST'
        )

        return response.object

    def ex_get_public_ip(self, ip_id):
        """
        Gets a Public IP

        :param ip_id: ID of the IP
        :type ip_id: ``str``

        :return: Instance of Public IP
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='public_ips/%s' % ip_id,
            method='GET'
        )

        return response.object

    def ex_delete_public_ip(self, ip_id):
        """
        Deletes a public IP

        :param ip_id: ID of public IP
        :type ip_id: ``str``

        :return: Instance of IP Address
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='public_ips/%s' % ip_id,
            method='DELETE'
        )

        return response

    def ex_update_public_ip(self, ip_id, reverse_dns):
        """
        Updates a Public IP

        :param ip_id: ID of public IP
        :type ip_id: ``str``

        :param reverse_dns: Reverse DNS
        :type reverse_dns: ``str``

        :return: Instance of Public IP
        :rtype: ``dict``
        """

        body = {
            'reverse_dns': reverse_dns
        }
        response = self.connection.request(
            action='public_ips/%s' % ip_id,
            data=body,
            method='DELETE'
        )

        return response.object

    """
    Private Network Operations
    """

    def ex_list_private_networks(self):
        """
        Lists all private networks

        :return: List of private networks
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='private_networks',
            method='GET'
        )

        return response.object

    def ex_create_private_network(self, name, description=None,
                                  datacenter_id=None,
                                  network_address=None,
                                  subnet_mask=None):
        """
        Creates a private network

        :param name: Name of the private network
        :type name: ``str``

        :param description: Description of the private network
        :type description: ``str``

        :param datacenter_id: ID of the data center for the private network
        :type datacenter_id: ``str``

        :param network_address: Network address of the private network
        :type network_address: ``str``

        :param subnet_mask: Subnet mask of the private network
        :type subnet_mask: ``str``

        :return: Newly created private network
        :rtype: ``dict``
        """

        body = {
            'name': name
        }

        if description is not None:
            body['description'] = description
        if datacenter_id is not None:
            body['datacenter_id'] = datacenter_id
        if network_address is not None:
            body['network_address'] = network_address
        if subnet_mask is not None:
            body['subnet_maks'] = subnet_mask

        response = self.connection.request(
            action='private_networks',
            data=body,
            method='POST'
        )
        return response.object

    def ex_delete_private_network(self, network_id):
        """
        Deletes a private network

        :param network_id: Id of the private network
        :type network_id: ``str``

        :return: Instance of the private network being deleted
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='private_networks' % network_id,
            method='DELETE'
        )

        return response.object

    def ex_update_private_network(self, network_id,
                                  name=None, description=None,
                                  datacenter_id=None,
                                  network_address=None,
                                  subnet_mask=None):
        """
       Updates a private network

       :param name: Name of the private network
       :type name: ``str``

       :param description: Description of the private network
       :type description: ``str``

       :param datacenter_id: ID of the data center for the private network
       :type datacenter_id: ``str``

       :param network_address: Network address of the private network
       :type network_address: ``str``

       :param subnet_mask: Subnet mask of the private network
       :type subnet_mask: ``str``

       :return: Instance of private network
       :rtype: ``dict``
       """
        body = {}

        if name is not None:
            body['name'] = name
        if description is not None:
            body['description'] = description
        if datacenter_id is not None:
            body['datacenter_id'] = datacenter_id
        if network_address is not None:
            body['network_address'] = network_address
        if subnet_mask is not None:
            body['subnet_maks'] = subnet_mask

        response = self.connection.request(
            action='private_networks/%s',
            data=body,
            method='PUT'
        )

        return response.object

    def ex_list_private_network_servers(self, network_id):
        """
        Lists all private network servers

        :param network_id: Private network ID
        :type network_id: ``str``

        :return: List of private network servers
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='/private_networks/%s/servers' % network_id,
            method='GET'
        )
        return response.object

    def ex_add_private_network_server(self, network_id, server_ids):
        """
        Add servers to private network

        :param network_id: Private Network ID
        :type network_id: ``str``

        :param server_ids: List of server IDs
        :type server_ids: ``list`` of ``str``

        :return: List of attached servers
        :rtype: ``dict``

        """
        body = {
            'servers': server_ids

        }

        response = self.connection.request(
            action='/private_networks/%s/servers' % network_id,
            data=body,
            method='POST'
        )

        return response.object

    def ex_remove_server_from_private_network(self, network_id, server_id):
        """
        Removes a server from the private network

        :param network_id: Private Network ID
        :type network_id: ``str``

        :param server_id: Id of the server
        :type server_id: ``str``

        :return: Instance of the private network
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='/private_networks/%s/servers/%s' % (network_id, server_id),
            method='POST'
        )
        return response.object

    """
    Monitoring policy operations
    """

    def ex_list_monitoring_policies(self):
        """
        Lists all monitoring policies

        :return: List of monitoring policies
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='monitoring_policies',
            method='GET'
        )

        return response.object

    def ex_create_monitoring_policy(self, name, thresholds,
                                    ports,
                                    processes,
                                    description=None,
                                    email=None,
                                    agent=None,
                                    ):
        """
        Creates a monitoring policy

        :param name: Name for the monitoring policy
        :type name: ``str``

        :param thresholds: Thresholds for the monitoring policy
        :type thresholds: ``dict``

        :param ports: Monitoring policies for ports
        :type ports: ``list`` of ``dict``

        :param processes: Processes to be monitored
        :type processes: ``list`` of ``dict``

        :param description: Description for the monitoring policy
        :type description: ``str``

        :param email: Email for notifications
        :type email: ``str``

        :param agent: Indicates if agent application will be installed
        :type agent: ``boolean``

        :return: Newly created instance of monitofing policy
        :rtype: ``dict``
        """
        body = {
            'name': name,
            'thresholds': thresholds,
            'ports': ports,
            'processes': processes
        }

        if description is not None:
            body['description'] = description
        if email is not None:
            body['email'] = email
        if agent is not None:
            body['agent'] = agent

        response = self.connection.request(
            action='monitoring_policies',
            data=body,
            method='POST'
        )
        return response.object

    def ex_delete_monitoring_policy(self, policy_id):
        """
        Deletes a monitoring policy

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :return: Instance of the monitoring policy being deleted
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='monitoring_policies' % policy_id,
            method='DELETE'
        )

        return response.object

    def ex_update_monitoring_policy(self, policy_id,
                                    email,
                                    thresholds,
                                    name=None, description=None):
        """
        Updates monitoring policy

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param email: Email to send notifications to
        :type email: ``str``

        :param thresholds: Thresholds for the monitoring policy
        :type thresholds: ``dict``

        :param name: Name of the monitoring policy
        :type name: ``str``

        :param description: Description of the monitoring policy
        :type description: ``str``

        :return: Instance of the monitoring policy being deleted
        :rtype: ``dict``
        """

        body = {}

        if name is not None:
            body['name'] = name
        if description is not None:
            body['description'] = description
        if thresholds is not None:
            body['thresholds'] = thresholds
        if email is not None:
            body['email'] = email

        response = self.connection.request(
            action='monitoring_policies/%s' % policy_id,
            data=body,
            method='PUT'
        )

        return response.object

    def ex_get_monitoring_policy(self, policy_id):
        """
        Fetches a monitoring policy

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='monitoring_policies/%s' % policy_id,
            method='GET'
        )

        return response.object

    def ex_get_monitoring_policy_ports(self, policy_id):
        """
        Fetches monitoring policy ports

        :param policy_id: Id of the monitoring policy
        :type policy_id:

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='monitoring_policies/%s/ports' % policy_id,
            method='GET'
        )

        return response.object

    def ex_get_monitoring_policy_port(self, policy_id, port_id):
        """
        Fetches monitoring policy port

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param port_id: Id of the port
        :type port_id: ``str``

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='monitoring_policies/%s/ports/%s' % (policy_id, port_id),
            method='GET'
        )

        return response.object

    def ex_remove_monitoring_policy_port(self, policy_id, port_id):
        """
        Removes monitoring policy port

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param port_id: Id of the port
        :type port_id: ``str``

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='monitoring_policies/%s/ports/%s' % (policy_id, port_id),
            method='DELETE'
        )

        return response.object

    def ex_add_monitoring_policy_ports(self, policy_id, ports):
        """
        Add monitoring policy ports

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param ports: List of ports
        :type ports: ``dict``
        [
           {
              'protocol':'TCP',
              'port':'80',
              'alert_if':'RESPONDING',
              'email_notification':true
           }
        ]

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """

        body = {'ports': ports}

        response = self.connection.request(
            action='monitoring_policies/%s/ports' % policy_id,
            data=body,
            method='POST'
        )

        return response.object

    def ex_get_monitoring_policy_processes(self, policy_id):
        """
        Fetches monitoring policy processes

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='monitoring_policies/%s/processes' % policy_id,
            method='GET'
        )

        return response.object

    def ex_get_monitoring_policy_process(self, policy_id, process_id):
        """
        Fetches monitoring policy process

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param process_id: Id of the process
        :type process_id: ``str``

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='monitoring_policies/%s/processes/%s'
                   % (policy_id, process_id),
            method='GET'
        )

        return response.object

    def ex_remove_monitoring_policy_process(self, policy_id, process_id):
        """
        Removes monitoring policy process

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param process_id: Id of the process
        :type process_id: ``str``

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """

        response = self.connection.request(
            action='monitoring_policies/%s/processes/%s'
                   % (policy_id, process_id),
            method='DELETE'
        )

        return response.object

    def ex_add_monitoring_policy_processes(self, policy_id, processes):
        """
        Add monitoring policy processes

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param processes: List of processes
        :type processes: ``list`` of ``dict``
        [
          {
            'process': 'taskmmgr',
            'alert_if': 'RUNNING',
            'email_notification': true
          }
        ]

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """

        body = {'processes': processes}

        response = self.connection.request(
            action='monitoring_policies/%s/processes' % policy_id,
            data=body,
            method='POST'
        )

        return response.object

    def ex_list_monitoring_policy_servers(self, policy_id):
        """
        List all servers that are being monitoried by the policy

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :return: List of servers being monitored
        :rtype: ``list`` of ``dict``
        """

        response = self.connection.request(
            action='monitoring_policies/%s/servers' % policy_id,
            method='GET'
        )

        return response.object

    def ex_add_servers_to_monitoring_policy(self, policy_id, servers):
        """
        Adds servers to monitoring policy

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param servers: List of server ID
        :type servers: ``list`` of ``str``

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """
        body = {
            'servers': servers
        }
        response = self.connection.request(
            action='monitoring_policies/%s/servers' % policy_id,
            data=body,
            method='POST'
        )

        return response.object

    def ex_remove_server_from_monitoring_policy(self, policy_id, server_id):
        """
        Removes a server from monitoring policy

        :param policy_id: Id of the monitoring policy
        :type policy_id: ``str``

        :param server_id: Id of the server
        :type server_id: ``str``

        :return: Instance of a monitoring policy
        :rtype: ``dict``
        """
        response = self.connection.request(
            action='monitoring_policies/%s/servers/%s'
                   % (policy_id, server_id),
            method='DELETE'
        )

        return response.object

    """
    Private Functions
    """

    def _to_images(self, object, image_type=None):
        if image_type is not None:
            images = [image for image in object if image['type'] == image_type]
        else:
            images = [image for image in object]

        return [self._to_image(image) for image in images]

    def _to_image(self, data):
        extra = {
            'os_family': data['os_family'],
            'os': data['os'],
            'os_version': data['os_version'],
            'os_architecture': data['os_architecture'],
            'os_image_type': data['os_image_type'],
            'min_hdd_size': data['min_hdd_size'],
            'available_datacenters': data['available_datacenters'],
            'licenses': data['licenses'],
            'version': data['version'],
            'categories': data['categories']
        }
        return NodeImage(id=data['id'], name=data['name'], driver=self,
                         extra=extra)

    def _to_node_size(self, data):
        return NodeSize(
            id=data['id'],
            name=data['name'],
            ram=data['hardware']['ram'],
            disk=data['hardware']['hdds'][0]['size'],
            bandwidth=None,
            price=None,
            driver=self.connection.driver,
            extra={
                'vcores': data['hardware']['vcore'],
                'cores_per_processor': data['hardware']['cores_per_processor']}

        )

    def _to_location(self, location):
        return NodeLocation(
            id=location['id'],
            name=location['country_code'],
            country=location['location'],
            driver=self.connection.driver
        )

    def _to_nodes(self, servers):
        return [self._to_node(
            server) for server in servers]

    def _to_node(self, server):
        extra = {}
        extra['datacenter'] = server['datacenter']

        if 'description' in server:
            extra['description'] = server['description']
        if 'status' in server:
            extra['status'] = server['status']
        if 'image' in server:
            extra['image'] = server['image']
        if 'hardware' in server:
            extra['hardware'] = server['hardware']
        if 'dvd' in server:
            extra['dvd'] = server['dvd']
        if 'snapshot' in server:
            extra['snapshot'] = server['snapshot']
        if 'ips' in server:
            extra['ips'] = server['ips']
        if 'alerts' in server:
            extra['alerts'] = server['alerts']
        if 'monitoring_policy' in server:
            extra['monitoring_policy'] = server['monitoring_policy']
        if 'private_networks' in server:
            extra['private_networks'] = server['private_networks']

        ips = []

        if server['ips'] is not None:
            for ip in server['ips']:
                ips.append(ip['ip'])
        state = self.NODE_STATE_MAP.get(
            server['status']['state'])

        return Node(
            id=server['id'],
            state=state,
            name=server['name'],
            driver=self.connection.driver,
            public_ips=ips,
            private_ips=None,
            extra=extra
        )

    def _wait_for_state(self, server_id, state, retries=50):
        for i in (0, retries):
            server = self.ex_get_server(server_id)

            if server.extra['status']['state'] == state:
                return
            sleep(5)

            if i == retries:
                raise Exception('Retries count reached')

    def _list_fixed_instances(self):
        response = self.connection.request(
            action='/servers/fixed_instance_sizes',
            method='GET'
        )

        return response.object
